@@ -1,23 +1,25 @@
+# frozen_string_literal: true
 $TESTING = true
-require 'rubygems'
+require 'bundler/setup'
 # require 'simplecov'
 # SimpleCov.start
 require 'minitest/pride' unless RUBY_ENGINE == 'rbx'
 require 'minitest/autorun'
 require 'mocha/setup'
-require 'memcached_mock'
+require_relative 'memcached_mock'
 
 ENV['MEMCACHED_SASL_PWDB'] = "#{File.dirname(__FILE__)}/sasl/sasldb"
 ENV['SASL_CONF_PATH'] = "#{File.dirname(__FILE__)}/sasl/memcached.conf"
 
-WANT_RAILS_VERSION = ENV['RAILS_VERSION'] || '>= 3.0.0'
-gem 'rails', WANT_RAILS_VERSION
 require 'rails'
 puts "Testing with Rails #{Rails.version}"
 
 require 'dalli'
 require 'logger'
 
+require 'active_support/time'
+require 'active_support/cache/dalli_store'
+
 Dalli.logger = Logger.new(STDOUT)
 Dalli.logger.level = Logger::ERROR
 
@@ -53,4 +55,9 @@ class MiniTest::Spec
     require 'action_controller'
     yield
   end
+
+  def with_connectionpool
+    require 'connection_pool'
+    yield
+  end
 end
