@@ -1,5 +1,5 @@
-require 'helper'
-require 'memcached_mock'
+# frozen_string_literal: true
+require_relative 'helper'
 
 describe 'Dalli' do
   describe 'options parsing' do
@@ -62,6 +62,16 @@ describe 'Dalli' do
     end
   end
 
+  describe 'ttl validation' do
+    it 'generated an ArgumentError for ttl that does not support to_i' do
+      memcached_persistent do |dc|
+        assert_raises ArgumentError do
+          dc.set('foo', 'bar', [])
+        end
+      end
+    end
+  end
+
   it "default to localhost:11211" do
     dc = Dalli::Client.new
     ring = dc.send(:ring)
@@ -134,7 +144,7 @@ describe 'Dalli' do
 
     it 'returns nil for nonexist key' do
       memcached_persistent do |dc|
-        assert_equal nil, dc.get('notexist')
+        assert_nil dc.get('notexist')
       end
     end
 
@@ -216,10 +226,25 @@ describe 'Dalli' do
         dc.set("fetch_key", false)
         res = dc.fetch("fetch_key") { flunk "fetch block called" }
         assert_equal false, res
+      end
+    end
+
+    it "support the fetch operation with nil values when cache_nils: true" do
+      memcached_persistent(21345, cache_nils: true) do |dc|
+        dc.flush
 
         dc.set("fetch_key", nil)
-        res = dc.fetch("fetch_key") { "bob" }
-        assert_equal 'bob', res
+        res = dc.fetch("fetch_key") { flunk "fetch block called" }
+        assert_nil res
+      end
+
+      memcached_persistent(21345, cache_nils: false) do |dc|
+        dc.flush
+        dc.set("fetch_key", nil)
+        executed = false
+        res = dc.fetch("fetch_key") { executed = true; 'bar' }
+        assert_equal 'bar', res
+        assert_equal true, executed
       end
     end
 
@@ -247,6 +272,22 @@ describe 'Dalli' do
       end
     end
 
+    it "support the cas! operation" do
+      memcached_persistent do |dc|
+        dc.flush
+
+        mutated = { 'blah' => 'foo!' }
+        resp = dc.cas!('cas_key') do |value|
+          assert_nil value
+          mutated
+        end
+        assert op_cas_succeeds(resp)
+
+        resp = dc.get('cas_key')
+        assert_equal mutated, resp
+      end
+    end
+
     it "support multi-get" do
       memcached_persistent do |dc|
         dc.close
@@ -427,6 +468,36 @@ describe 'Dalli' do
       end
     end
 
+    it 'allow TCP connections to configure SO_RCVBUF' do
+      memcached_persistent do |dc, port|
+        value = 5000
+        dc = Dalli::Client.new("localhost:#{port}", :rcvbuf => value)
+        dc.set(:a, 1)
+        ring = dc.send(:ring)
+        server = ring.servers.first
+        socket = server.instance_variable_get('@sock')
+
+        optval = socket.getsockopt(Socket::SOL_SOCKET, Socket::SO_RCVBUF)
+        expected = jruby? ? value : value * 2
+        assert_equal expected, optval.unpack('i')[0]
+      end
+    end
+
+    it 'allow TCP connections to configure SO_SNDBUF' do
+      memcached_persistent do |dc, port|
+        value = 5000
+        dc = Dalli::Client.new("localhost:#{port}", :sndbuf => value)
+        dc.set(:a, 1)
+        ring = dc.send(:ring)
+        server = ring.servers.first
+        socket = server.instance_variable_get('@sock')
+
+        optval = socket.getsockopt(Socket::SOL_SOCKET, Socket::SO_SNDBUF)
+        expected = jruby? ? value : value * 2
+        assert_equal expected, optval.unpack('i')[0]
+      end
+    end
+
     it "pass a simple smoke test" do
       memcached_persistent do |dc, port|
         resp = dc.flush
@@ -437,7 +508,7 @@ describe 'Dalli' do
         assert_equal 'bar', dc.get(:foo)
 
         resp = dc.get('123')
-        assert_equal nil, resp
+        assert_nil resp
 
         assert op_addset_succeeds(dc.set('123', 'xyz'))
 
@@ -491,7 +562,7 @@ describe 'Dalli' do
         assert_equal 'bar', dc.get(:foo)
 
         resp = dc.get('123')
-        assert_equal nil, resp
+        assert_nil resp
 
         assert op_addset_succeeds(dc.set('123', 'xyz'))
 
@@ -615,6 +686,16 @@ describe 'Dalli' do
       end
     end
 
+    it 'handle special Regexp characters in namespace with get_multi' do
+      memcached_persistent do |dc, port|
+        # /(?!)/ is a contradictory PCRE and should never be able to match
+        dc = Dalli::Client.new("localhost:#{port}", :namespace => '(?!)')
+        dc.set('a', 1)
+        dc.set('b', 2)
+        assert_equal({'a' => 1, 'b' => 2}, dc.get_multi('a', 'b'))
+      end
+    end
+
     it "handle application marshalling issues" do
       memcached_persistent do |dc|
         old = Dalli.logger
@@ -637,6 +718,13 @@ describe 'Dalli' do
           dalli.set('verylarge', value)
         end
       end
+
+      it 'allow large values to be set' do
+        memcached_persistent do |dc|
+          value = "0"*1024*1024
+          assert dc.set('verylarge', value, nil, :compress => true)
+        end
+      end
     end
 
     describe 'in low memory conditions' do
