@@ -1,9 +1,8 @@
 # encoding: utf-8
-require 'helper'
-require 'memcached_mock'
+# frozen_string_literal: true
+require_relative 'helper'
 
 describe 'Encoding' do
-
   describe 'using a live server' do
     it 'support i18n content' do
       memcached_persistent do |dc|
@@ -24,9 +23,8 @@ describe 'Encoding' do
         assert dc.set(key, 'bar', 1)
         assert_equal 'bar', dc.get(key)
         sleep 1.2
-        assert_equal nil, dc.get(key)
+        assert_nil dc.get(key)
       end
     end
-
   end
 end
