@@ -1,10 +1,9 @@
 # encoding: utf-8
-require 'helper'
+# frozen_string_literal: true
+require_relative 'helper'
 require 'json'
-require 'memcached_mock'
 
 describe 'Serializer' do
-
   it 'default to Marshal' do
     memcached(29198) do |dc|
       dc.set 1,2
